@@ -1,7 +1,5 @@
 """Selector is a single Selector of a CSSStyleRule SelectorList.
-
-Partly implements
-    http://www.w3.org/TR/css3-selectors/
+Partly implements http://www.w3.org/TR/css3-selectors/.
 
 TODO
     - .contains(selector)
@@ -9,45 +7,19 @@ TODO
 """
 __all__ = ['Selector']
 __docformat__ = 'restructuredtext'
-__version__ = '$Id: selector.py 1352 2008-07-12 11:01:04Z cthedot $'
+__version__ = '$Id$'
 
-import xml.dom
-import cssutils
+from cssutils.helper import Deprecated
 from cssutils.util import _SimpleNamespaces
+import cssutils
+import xml.dom
 
 class Selector(cssutils.util.Base2):
     """
-    (cssutils) a single selector in a SelectorList of a CSSStyleRule
+    (cssutils) a single selector in a :class:`~cssutils.css.SelectorList` 
+    of a :class:`~cssutils.css.CSSStyleRule`.
 
-    Properties
-    ==========
-    element
-        Effective element target of this selector
-    parentList: of type SelectorList, readonly
-        The SelectorList that contains this selector or None if this
-        Selector is not attached to a SelectorList.
-    selectorText
-        textual representation of this Selector
-    seq
-        sequence of Selector parts including comments
-    specificity (READONLY)
-        tuple of (a, b, c, d) where:
-        
-        a
-            presence of style in document, always 0 if not used on a document
-        b
-            number of ID selectors
-        c 
-            number of .class selectors
-        d 
-            number of Element (type) selectors
-        
-    wellformed
-        if this selector is wellformed regarding the Selector spec
-
-    Format
-    ======
-    ::
+    Format::
 
         # implemented in SelectorList
         selectors_group
@@ -127,22 +99,22 @@ class Selector(cssutils.util.Base2):
           ;
 
     """
-    def __init__(self, selectorText=None, parentList=None,
+    def __init__(self, selectorText=None, parent=None,
                  readonly=False):
         """
         :Parameters:
             selectorText
                 initial value of this selector
-            parentList
+            parent
                 a SelectorList
             readonly
                 default to False
         """
         super(Selector, self).__init__()
 
-        self.__namespaces = _SimpleNamespaces()
+        self.__namespaces = _SimpleNamespaces(log=self._log)
         self._element = None
-        self._parent = parentList
+        self._parent = parent
         self._specificity = (0, 0, 0, 0)
         
         if selectorText:
@@ -150,30 +122,67 @@ class Selector(cssutils.util.Base2):
 
         self._readonly = readonly
 
+    def __repr__(self):
+        if self.__getNamespaces():
+            st = (self.selectorText, self._getUsedNamespaces())
+        else:
+            st = self.selectorText
+        return u"cssutils.css.%s(selectorText=%r)" % (self.__class__.__name__, 
+                                                      st)
+
+    def __str__(self):
+        return u"<cssutils.css.%s object selectorText=%r specificity=%r" \
+               u" _namespaces=%r at 0x%x>" % (self.__class__.__name__,
+                                              self.selectorText,
+                                              self.specificity,
+                                              self._getUsedNamespaces(),
+                                              id(self))
+
+    def _getUsedUris(self):
+        "Return list of actually used URIs in this Selector."
+        uris = set()
+        for item in self.seq:
+            type_, val = item.type, item.value
+            if type_.endswith(u'-selector') or type_ == u'universal' and \
+               isinstance(val, tuple) and val[0] not in (None, u'*'):
+                uris.add(val[0])
+        return uris
+
+    def _getUsedNamespaces(self):
+        "Return actually used namespaces only."
+        useduris = self._getUsedUris()
+        namespaces = _SimpleNamespaces(log=self._log)
+        for p, uri in self._namespaces.items():
+            if uri in useduris:
+                namespaces[p] = uri
+        return namespaces
+
     def __getNamespaces(self):
-        "uses own namespaces if not attached to a sheet, else the sheet's ones"
+        "Use own namespaces if not attached to a sheet, else the sheet's ones."
         try:
             return self._parent.parentRule.parentStyleSheet.namespaces
         except AttributeError:
             return self.__namespaces
 
-    _namespaces = property(__getNamespaces, doc="""if this Selector is attached
-        to a CSSStyleSheet the namespaces of that sheet are mirrored here.
-        While the Selector (or parent SelectorList or parentRule(s) of that are
-        not attached a own dict of {prefix: namespaceURI} is used.""")
+    _namespaces = property(__getNamespaces, 
+                           doc=u"If this Selector is attached to a "
+                               u"CSSStyleSheet the namespaces of that sheet "
+                               u"are mirrored here. While the Selector (or "
+                               u"parent SelectorList or parentRule(s) of that "
+                               u"are not attached a own dict of {prefix: "
+                               u"namespaceURI} is used.")
 
     
     element = property(lambda self: self._element, 
                        doc=u"Effective element target of this selector.")
 
-    parentList = property(lambda self: self._parent,
-        doc="(DOM) The SelectorList that contains this Selector or\
-        None if this Selector is not attached to a SelectorList.")
-        
+    parent = property(lambda self: self._parent,
+                      doc=u"(DOM) The SelectorList that contains this Selector "
+                          u"or None if this Selector is not attached to a "
+                          u"SelectorList.")
+                
     def _getSelectorText(self):
-        """
-        returns serialized format
-        """
+        """Return serialized format."""
         return cssutils.ser.do_css_Selector(self)
 
     def _setSelectorText(self, selectorText):
@@ -183,23 +192,25 @@ class Selector(cssutils.util.Base2):
             Given namespaces are ignored if this object is attached to a 
             CSSStyleSheet!
         
-        :Exceptions:
-            - `NAMESPACE_ERR`: (self)
+        :exceptions:
+            - :exc:`~xml.dom.NamespaceErr`:
               Raised if the specified selector uses an unknown namespace
               prefix.
-            - `SYNTAX_ERR`: (self)
+            - :exc:`~xml.dom.SyntaxErr`:
               Raised if the specified CSS string value has a syntax error
               and is unparsable.
-            - `NO_MODIFICATION_ALLOWED_ERR`: (self)
+            - :exc:`~xml.dom.NoModificationAllowedErr`:
               Raised if this rule is readonly.
         """
         self._checkReadonly()
         
         # might be (selectorText, namespaces)
         selectorText, namespaces = self._splitNamespacesOff(selectorText)
+
         try:
-            # uses parent stylesheets namespaces if available, otherwise given ones
-            namespaces = self.parentList.parentRule.parentStyleSheet.namespaces
+            # uses parent stylesheets namespaces if available, 
+            # otherwise given ones
+            namespaces = self.parent.parentRule.parentStyleSheet.namespaces
         except AttributeError:
             pass
         tokenizer = self._tokenize2(selectorText)
@@ -259,8 +270,8 @@ class Selector(cssutils.util.Base2):
                     tokens.append(('universal', val, lin, col))
 
                 elif val == u'|' and tokens and\
-                     self._type(tokens[-1]) in (self._prods.IDENT, 'universal') and\
-                     self._tokenvalue(tokens[-1]).find(u'|') == -1:
+                     self._type(tokens[-1]) in (self._prods.IDENT, 'universal')\
+                     and self._tokenvalue(tokens[-1]).find(u'|') == -1:
                     # namespace_prefix: "IDENT|" or "*|"
                     tokens[-1] = ('namespace_prefix', 
                                   self._tokenvalue(tokens[-1])+u'|', lin, col)
@@ -325,22 +336,25 @@ class Selector(cssutils.util.Base2):
                         # *|name: in ANY_NS
                         namespaceURI = cssutils._ANYNS
                     elif prefix is None:
-                        # e or *: default namespace with prefix u'' or local-name()
+                        # e or *: default namespace with prefix u'' 
+                        # or local-name()
                         namespaceURI = namespaces.get(u'', None)
                     elif prefix == u'':
                         # |name or |*: in no (or the empty) namespace
                         namespaceURI = u''
                     else:
                         # explicit namespace prefix
-                        try:
-                            namespaceURI = namespaces[prefix]
-                        except KeyError:
+                        # does not raise KeyError, see _SimpleNamespaces
+                        namespaceURI = namespaces[prefix]
+
+                        if namespaceURI is None:
                             new['wellformed'] = False
-                            self._log.error(
-                                u'Selector: No namespaceURI found for prefix %r' %
-                                prefix, token=token, error=xml.dom.NamespaceErr)
+                            self._log.error(u'Selector: No namespaceURI found '
+                                            u'for prefix %r' % prefix, 
+                                            token=token, 
+                                            error=xml.dom.NamespaceErr)
                             return
-                    
+                            
                     # val is now (namespaceprefix, name) tuple
                     val = (namespaceURI, val)
 
@@ -360,7 +374,8 @@ class Selector(cssutils.util.Base2):
                 seq.append(val, typ, line=line, col=col)
 
             # expected constants
-            simple_selector_sequence = 'type_selector universal HASH class attrib pseudo negation '
+            simple_selector_sequence = 'type_selector universal HASH class ' \
+                                       'attrib pseudo negation '
             simple_selector_sequence2 = 'HASH class attrib pseudo negation '
 
             element_name = 'element_name'
@@ -442,21 +457,28 @@ class Selector(cssutils.util.Base2):
                 # pseudo-class or pseudo-element :a ::a :a( ::a(
                 """
                 /* '::' starts a pseudo-element, ':' a pseudo-class */
-                /* Exceptions: :first-line, :first-letter, :before and :after. */
-                /* Note that pseudo-elements are restricted to one per selector and */
+                /* Exceptions: :first-line, :first-letter, :before and 
+                :after. */
+                /* Note that pseudo-elements are restricted to one per selector
+                and */
                 /* occur only in the last simple_selector_sequence. */
                 """
                 context = new['context'][-1]
-                val, typ = self._tokenvalue(token, normalize=True), self._type(token)
+                val, typ = self._tokenvalue(token, normalize=True),\
+                           self._type(token)
                 if 'pseudo' in expected:                    
-                    if val in (':first-line', ':first-letter', ':before', ':after'):
+                    if val in (':first-line',
+                               ':first-letter',
+                               ':before',
+                               ':after'):
                         # always pseudo-element ???
                         typ = 'pseudo-element'
                     append(seq, val, typ, token=token)
                     
                     if val.endswith(u'('):
                         # function
-                        new['context'].append(typ) # "pseudo-" "class" or "element"
+                        # "pseudo-" "class" or "element"
+                        new['context'].append(typ) 
                         return expressionstart                 
                     elif 'negation' == context:
                         return negationend
@@ -559,9 +581,7 @@ class Selector(cssutils.util.Base2):
 
                 else:
                     new['wellformed'] = False
-                    self._log.error(
-                        u'Selector: Unexpected IDENT.', 
-                        token=token)
+                    self._log.error(u'Selector: Unexpected IDENT.', token=token)
                     return expected
 
             def _class(expected, seq, token, tokenizer=None):
@@ -578,8 +598,7 @@ class Selector(cssutils.util.Base2):
 
                 else:
                     new['wellformed'] = False
-                    self._log.error(
-                        u'Selector: Unexpected class.', token=token)
+                    self._log.error(u'Selector: Unexpected class.', token=token)
                     return expected
 
             def _hash(expected, seq, token, tokenizer=None):
@@ -596,8 +615,7 @@ class Selector(cssutils.util.Base2):
 
                 else:
                     new['wellformed'] = False
-                    self._log.error(
-                        u'Selector: Unexpected HASH.', token=token)
+                    self._log.error(u'Selector: Unexpected HASH.', token=token)
                     return expected
 
             def _char(expected, seq, token, tokenizer=None):
@@ -616,7 +634,8 @@ class Selector(cssutils.util.Base2):
                     else:
                         return simple_selector_sequence2 + combinator
 
-                elif u'=' == val and 'attrib' == context and 'combinator' in expected:
+                elif u'=' == val and 'attrib' == context\
+                     and 'combinator' in expected:
                     # combinator in attrib
                     append(seq, val, 'equals', token=token)
                     return attvalue
@@ -676,6 +695,7 @@ class Selector(cssutils.util.Base2):
                         u'Selector: Single selector only.', 
                         error=xml.dom.InvalidModificationErr, 
                         token=token)
+                    return expected
 
                 else:
                     new['wellformed'] = False
@@ -697,10 +717,19 @@ class Selector(cssutils.util.Base2):
                         u'Selector: Unexpected negation.', token=token)
                     return expected
 
+            def _atkeyword(expected, seq, token, tokenizer=None):
+                "invalidates selector"
+                new['wellformed'] = False
+                self._log.error(
+                        u'Selector: Unexpected ATKEYWORD.', token=token)
+                return expected
+
+
             # expected: only|not or mediatype, mediatype, feature, and
             newseq = self._tempSeq()
             
-            wellformed, expected = self._parse(expected=simple_selector_sequence,
+            wellformed, expected = self._parse(
+                expected=simple_selector_sequence,
                 seq=newseq, tokenizer=tokenizer,
                 productions={'CHAR': _char,
                              'class': _class,
@@ -723,27 +752,28 @@ class Selector(cssutils.util.Base2):
                              'INCLUDES': _attcombinator,
                              
                              'S': _S,
-                             'COMMENT': _COMMENT})
+                             'COMMENT': _COMMENT,
+                             'ATKEYWORD': _atkeyword})
             wellformed = wellformed and new['wellformed']
 
             # post condition         
             if len(new['context']) > 1 or not newseq:
                 wellformed = False
-                self._log.error(u'Selector: Invalid or incomplete selector: %s' %
-                    self._valuestr(selectorText))
+                self._log.error(u'Selector: Invalid or incomplete selector: %s' 
+                                % self._valuestr(selectorText))
             
             if expected == 'element_name':
                 wellformed = False
-                self._log.error(u'Selector: No element name found: %s' %
-                    self._valuestr(selectorText))
+                self._log.error(u'Selector: No element name found: %s'
+                                % self._valuestr(selectorText))
 
             if expected == simple_selector_sequence and newseq:
                 wellformed = False
-                self._log.error(u'Selector: Cannot end with combinator: %s' %
-                    self._valuestr(selectorText))
+                self._log.error(u'Selector: Cannot end with combinator: %s'
+                                % self._valuestr(selectorText))
 
-            if newseq and hasattr(newseq[-1].value, 'strip') and \
-               newseq[-1].value.strip() == u'':
+            if newseq and hasattr(newseq[-1].value, 'strip') \
+               and newseq[-1].value.strip() == u'':
                 del newseq[-1]
 
             # set
@@ -756,42 +786,29 @@ class Selector(cssutils.util.Base2):
                 self.__namespaces = self._getUsedNamespaces()
 
     selectorText = property(_getSelectorText, _setSelectorText,
-        doc="(DOM) The parsable textual representation of the selector.")
-
+                            doc=u"(DOM) The parsable textual representation of "
+                                u"the selector.")
 
     specificity = property(lambda self: self._specificity, 
-                           doc="Specificity of this selector (READONLY).")
+         doc="""Specificity of this selector (READONLY). 
+                Tuple of (a, b, c, d) where: 
+                
+                a
+                    presence of style in document, always 0 if not used on a 
+                    document
+                b
+                    number of ID selectors
+                c 
+                    number of .class selectors
+                d 
+                    number of Element (type) selectors""")
 
     wellformed = property(lambda self: bool(len(self.seq)))
 
-    def __repr__(self):
-        if self.__getNamespaces():
-            st = (self.selectorText, self._getUsedNamespaces())
-        else:
-            st = self.selectorText
-        return u"cssutils.css.%s(selectorText=%r)" % (
-                self.__class__.__name__, st)
 
-    def __str__(self):
-        return u"<cssutils.css.%s object selectorText=%r specificity=%r _namespaces=%r at 0x%x>" % (
-                self.__class__.__name__, self.selectorText, self.specificity, 
-                self._getUsedNamespaces(), id(self))
-
-    def _getUsedUris(self):
-        "returns list of actually used URIs in this Selector"
-        uris = set()
-        for item in self.seq:
-            type_, val = item.type, item.value
-            if type_.endswith(u'-selector') or type_ == u'universal' and \
-               type(val) == tuple and val[0] not in (None, u'*'):
-                uris.add(val[0])
-        return uris
-
-    def _getUsedNamespaces(self):
-        "returns actually used namespaces only"
-        useduris = self._getUsedUris()
-        namespaces = _SimpleNamespaces()
-        for p, uri in self._namespaces.items():
-            if uri in useduris:
-                namespaces[p] = uri
-        return namespaces
+    @Deprecated('Use property parent instead')
+    def _getParentList(self):
+        return self.parent
+    
+    parentList = property(_getParentList,
+                          doc="DEPRECATED, see property parent instead")
