@@ -1,33 +1,33 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-"""serializer classes for CSS classes
-
-"""
+"""cssutils serializer"""
 __all__ = ['CSSSerializer', 'Preferences']
 __docformat__ = 'restructuredtext'
-__version__ = '$Id: serialize.py 1329 2008-07-09 12:28:02Z cthedot $'
+__version__ = '$Id$'
+
+from cssutils.helper import normalize
 import codecs
-import re
 import cssutils
-import util
+import helper
+import re
+import xml.dom
 
 def _escapecss(e):
     """
     Escapes characters not allowed in the current encoding the CSS way
     with a backslash followed by a uppercase hex code point
-    
+
     E.g. the german umlaut 'ä' is escaped as \E4
     """
     s = e.object[e.start:e.end]
     return u''.join([ur'\%s ' % str(hex(ord(x)))[2:] # remove 0x from hex
                      .upper() for x in s]), e.end
 
-codecs.register_error('escapecss', _escapecss)               
+codecs.register_error('escapecss', _escapecss)
 
 
 class Preferences(object):
-    """
-    controls output of CSSSerializer
+    """Control output of CSSSerializer.
 
     defaultAtKeyword = True
         Should the literal @keyword from src CSS be used or the default
@@ -37,20 +37,24 @@ class Preferences(object):
         the src file, e.g. if ``True``: ``color`` else: ``c\olor``
 
         Only used if ``keepAllProperties==False``.
-        
+
     defaultPropertyPriority = True
-        Should the normalized or literal priority be used, e.g. '!important'
-        or u'!Im\portant'
+        Should the normalized or literal priority be used, e.g. ``!important``
+        or ``!Im\portant``
 
     importHrefFormat = None
-        Uses hreftype if ``None`` or explicit ``'string'`` or ``'uri'``
+        Uses hreftype if ``None`` or format ``"URI"`` if ``'string'`` or
+        format ``url(URI)`` if ``'uri'``
     indent = 4 * ' '
         Indentation of e.g Properties inside a CSSStyleDeclaration
-    indentSpecificities = False
+    indentClosingBrace = True
+        Defines if closing brace of block is indented to match indentation 
+        of the block (default) oder match indentation of selector.
+    indentSpecificities = False (**EXPERIMENTAL**)
         Indent rules with subset of Selectors and higher Specitivity
-        
+
     keepAllProperties = True
-        If ``True`` all properties set in the original CSSStylesheet 
+        If ``True`` all properties set in the original CSSStylesheet
         are kept meaning even properties set twice with the exact same
         same name are kept!
     keepComments = True
@@ -58,131 +62,141 @@ class Preferences(object):
     keepEmptyRules = False
         defines if empty rules like e.g. ``a {}`` are kept in the resulting
         serialized sheet
+    keepUnknownAtRules = True
+        defines if unknown @rules like e.g. ``@three-dee {}`` are kept in the
+        serialized sheet
     keepUsedNamespaceRulesOnly = False
         if True only namespace rules which are actually used are kept
-        
+
     lineNumbers = False
         Only used if a complete CSSStyleSheet is serialized.
     lineSeparator = u'\\n'
-        How to end a line. This may be set to e.g. u'' for serializing of 
+        How to end a line. This may be set to e.g. u'' for serializing of
         CSSStyleDeclarations usable in HTML style attribute.
     listItemSpacer = u' '
         string which is used in ``css.SelectorList``, ``css.CSSValue`` and
         ``stylesheets.MediaList`` after the comma
+    normalizedVarNames = True
+        defines if variable names should be serialized normalized (they are
+        used as being normalized anyway)
     omitLastSemicolon = True
         If ``True`` omits ; after last property of CSSStyleDeclaration
+    omitLeadingZero = False
+        defines if values between -1 and 1 should omit the 0, like ``.5px``
     paranthesisSpacer = u' '
-        string which is used before an opening paranthesis like in a 
+        string which is used before an opening paranthesis like in a
         ``css.CSSMediaRule`` or ``css.CSSStyleRule``
     propertyNameSpacer = u' '
         string which is used after a Property name colon
+    resolveVariables = True
+        if ``True`` all variable references are tried to resolved and
+        all CSSVariablesRules are removed from the output.
+        Any variable reference not resolvable is simply kept untouched.
     selectorCombinatorSpacer = u' '
         string which is used before and after a Selector combinator like +, > or ~.
         CSSOM defines a single space for this which is also the default in cssutils.
     spacer = u' '
         general spacer, used e.g. by CSSUnknownRule
-    
-    validOnly = False **DO NOT CHANGE YET**
-        if True only valid (currently Properties) are kept
-        
+
+    validOnly = False
+        if True only valid (Properties) are output
+
         A Property is valid if it is a known Property with a valid value.
-        Currently CSS 2.1 values as defined in cssproperties.py would be
-        valid.
-        
     """
     def __init__(self, **initials):
-        """
-        Always use named instead of positional parameters
-        """
+        """Always use named instead of positional parameters."""
         self.useDefaults()
-        
         for key, value in initials.items():
             if value:
                 self.__setattr__(key, value)
 
+    def __repr__(self):
+        return u"cssutils.css.%s(%s)" % (self.__class__.__name__,
+            u', '.join(['\n    %s=%r' % (p, self.__getattribute__(p)) for p in self.__dict__]
+                ))
+
+    def __str__(self):
+        return u"<cssutils.css.%s object %s at 0x%x" % (self.__class__.__name__,
+            u' '.join(['%s=%r' % (p, self.__getattribute__(p)) for p in self.__dict__]
+                ),
+                id(self))
+
     def useDefaults(self):
-        "reset all preference options to the default value"
+        "Reset all preference options to their default value."
         self.defaultAtKeyword = True
         self.defaultPropertyName = True
         self.defaultPropertyPriority = True
         self.importHrefFormat = None
         self.indent = 4 * u' '
+        self.indentClosingBrace = True
         self.indentSpecificities = False
         self.keepAllProperties = True
         self.keepComments = True
         self.keepEmptyRules = False
+        self.keepUnknownAtRules = True
         self.keepUsedNamespaceRulesOnly = False
         self.lineNumbers = False
         self.lineSeparator = u'\n'
         self.listItemSpacer = u' '
+        self.normalizedVarNames = True
         self.omitLastSemicolon = True
+        self.omitLeadingZero = False
         self.paranthesisSpacer = u' '
         self.propertyNameSpacer = u' '
+        self.resolveVariables = True
         self.selectorCombinatorSpacer = u' '
         self.spacer = u' '
         self.validOnly = False # should not be changed currently!!!
-        
+
     def useMinified(self):
-        """
-        sets options to achive a minified stylesheet
-        
-        you may want to set preferences with this convenience method
-        and set settings you want adjusted afterwards
+        """Set options resulting in a minified stylesheet.
+
+        You may want to set preferences with this convenience method
+        and set settings you want adjusted afterwards.
         """
         self.importHrefFormat = 'string'
         self.indent = u''
         self.keepComments = False
         self.keepEmptyRules = False
+        self.keepUnknownAtRules = False
         self.keepUsedNamespaceRulesOnly = True
         self.lineNumbers = False
         self.lineSeparator = u''
         self.listItemSpacer = u''
         self.omitLastSemicolon = True
+        self.omitLeadingZero = True
         self.paranthesisSpacer = u''
         self.propertyNameSpacer = u''
         self.selectorCombinatorSpacer = u''
         self.spacer = u''
         self.validOnly = False
 
-    def __repr__(self):
-        return u"cssutils.css.%s(%s)" % (self.__class__.__name__, 
-            u', '.join(['\n    %s=%r' % (p, self.__getattribute__(p)) for p in self.__dict__]
-                ))
-
-    def __str__(self):
-        return u"<cssutils.css.%s object %s at 0x%x" % (self.__class__.__name__, 
-            u' '.join(['%s=%r' % (p, self.__getattribute__(p)) for p in self.__dict__]
-                ),
-                id(self))
-
 
 class Out(object):
-    """
-    a simple class which makes appended items available as a combined string
-    """
+    """A simple class which makes appended items available as a combined string"""
     def __init__(self, ser):
         self.ser = ser
         self.out = []
-    
+
     def _remove_last_if_S(self):
         if self.out and not self.out[-1].strip():
             # remove trailing S
             del self.out[-1]
-    
-    def append(self, val, typ=None, space=True, keepS=False, indent=False, 
+
+    def append(self, val, typ=None, space=True, keepS=False, indent=False,
                lineSeparator=False):
         """Appends val. Adds a single S after each token except as follows:
-        
+
         - typ COMMENT
             uses cssText depending on self.ser.prefs.keepComments
         - typ "Property", cssutils.css.CSSRule.UNKNOWN_RULE
             uses cssText
         - typ STRING
-            escapes ser._string
+            escapes helper.string
         - typ S
             ignored except ``keepS=True``
         - typ URI
-            calls ser_uri
+            calls helper.uri
         - val ``{``
             adds LF after
         - val ``;``
@@ -194,32 +208,49 @@ class Out(object):
         - some other vals
             add ``*spacer`` except ``space=False``
         """
-        if val or 'STRING' == typ:
+        prefspace = self.ser.prefs.spacer
+        if val or typ in ('STRING', 'URI'):
             # PRE
             if 'COMMENT' == typ:
                 if self.ser.prefs.keepComments:
                     val = val.cssText
-                else: 
+                else:
                     return
-            elif typ in ('Property', cssutils.css.CSSRule.UNKNOWN_RULE):
+            elif hasattr(val, 'cssText'):
                 val = val.cssText
+#            elif typ in ('Property', cssutils.css.CSSRule.UNKNOWN_RULE):
+#                val = val.cssText
             elif 'S' == typ and not keepS:
                 return
+            elif 'S' == typ and keepS:
+                val = u' '
+#            elif typ in ('NUMBER', 'DIMENSION', 'PERCENTAGE') and val == u'0':
+#                # remove sign + or - if value is zero
+#                # TODO: only for lenghts!
+#                if self.out and self.out[-1] in u'+-':
+#                    del self.out[-1]
             elif 'STRING' == typ:
                 # may be empty but MUST not be None
-                if val is None: 
+                if val is None:
                     return
-                val = self.ser._string(val)                
+                val = helper.string(val)
+                if not prefspace:
+                    self._remove_last_if_S()
             elif 'URI' == typ:
-                val = self.ser._uri(val)
-            elif val in u'+>~,:{;)]':
+                val = helper.uri(val)
+            elif 'HASH' == typ:
+                val = self.ser._hash(val)
+            elif val in u'+>~,:{;)]/=':
                 self._remove_last_if_S()
-                
+
             # APPEND
             if indent:
                 self.out.append(self.ser._indentblock(val, self.ser._level+1))
             else:
+                if val.endswith(u' '):
+                    self._remove_last_if_S()
                 self.out.append(val)
+
             # POST
             if lineSeparator:
                 # Property , ...
@@ -227,6 +258,9 @@ class Out(object):
             elif val in u'+>~': # enclose selector combinator
                 self.out.insert(-1, self.ser.prefs.selectorCombinatorSpacer)
                 self.out.append(self.ser.prefs.selectorCombinatorSpacer)
+            elif u')' == val and not keepS: # CHAR funcend
+                # TODO: pref?
+                self.out.append(u' ')
             elif u',' == val: # list
                 self.out.append(self.ser.prefs.listItemSpacer)
             elif u':' == val: # prop
@@ -236,37 +270,37 @@ class Out(object):
                 self.out.append(self.ser.prefs.lineSeparator)
             elif u';' == val: # end or prop or block
                 self.out.append(self.ser.prefs.lineSeparator)
-            elif val not in u'}[]()' and space:
+            elif val not in u'}[]()/=' and typ != 'FUNCTION' and space:
                 self.out.append(self.ser.prefs.spacer)
-        
-    def value(self, delim=u'', end=None):
+                if typ != 'STRING' and not self.ser.prefs.spacer and \
+                   self.out and not self.out[-1].endswith(u' '):
+                    self.out.append(u' ')
+
+    def value(self, delim=u'', end=None, keepS=False):
         "returns all items joined by delim"
-        self._remove_last_if_S()
+        if not keepS:
+            self._remove_last_if_S()
         if end:
             self.out.append(end)
         return delim.join(self.out)
-    
+
 
 class CSSSerializer(object):
-    """
-    Methods to serialize a CSSStylesheet and its parts
+    """Serialize a CSSStylesheet and its parts.
 
     To use your own serializing method the easiest is to subclass CSS
     Serializer and overwrite the methods you like to customize.
     """
-    # chars not in URI without quotes around
-    __forbidden_in_uri_matcher = re.compile(ur'''.*?[\)\s\;]''', re.U).match
-
     def __init__(self, prefs=None):
         """
-        prefs
+        :param prefs:
             instance of Preferences
         """
         if not prefs:
             prefs = Preferences()
         self.prefs = prefs
         self._level = 0 # current nesting level
-        
+
         # TODO:
         self._selectors = [] # holds SelectorList
         self._selectorlevel = 0 # current specificity nesting level
@@ -310,29 +344,23 @@ class CSSSerializer(object):
             text = self.prefs.lineSeparator.join(out)
         return text
 
-    def _string(self, s):
+    def _hash(self, val, type_=None):
         """
-        returns s encloded between "..." and escaped delim charater ", 
-        escape line breaks \\n \\r and \\f
+        Short form of hash, e.g. #123 instead of #112233
         """
-        # \n = 0xa, \r = 0xd, \f = 0xc
-        s = s.replace('\n', '\\a ').replace(
-                      '\r', '\\d ').replace(
-                      '\f', '\\c ')
-        return u'"%s"' % s.replace('"', u'\\"')
-
-    def _uri(self, uri):
-        """returns uri enclosed in url() and "..." if necessary"""
-        if CSSSerializer.__forbidden_in_uri_matcher(uri):
-            return 'url(%s)' % self._string(uri)
+        # TODO: add pref for this!
+        if len(val) == 7 and val[1] == val[2] and\
+                             val[3] == val[4] and\
+                             val[5] == val[6]:
+            return u'#%s%s%s' % (val[1], val[3], val[5])
         else:
-            return 'url(%s)' % uri
+            return val
 
     def _valid(self, x):
         "checks items valid property and prefs.validOnly"
-        return not self.prefs.validOnly or (self.prefs.validOnly and 
+        return not self.prefs.validOnly or (self.prefs.validOnly and
                                             x.valid)
-    
+
     def do_CSSStyleSheet(self, stylesheet):
         """serializes a complete CSSStyleSheet"""
         useduris = stylesheet._getUsedURIs()
@@ -342,20 +370,21 @@ class CSSSerializer(object):
                rule.NAMESPACE_RULE == rule.type and\
                rule.namespaceURI not in useduris and (
                     rule.prefix or None not in useduris):
-                continue 
+                continue
 
             cssText = rule.cssText
             if cssText:
                 out.append(cssText)
         text = self._linenumnbers(self.prefs.lineSeparator.join(out))
-        
+
         # get encoding of sheet, defaults to UTF-8
         try:
             encoding = stylesheet.cssRules[0].encoding
         except (IndexError, AttributeError):
-            encoding = 'UTF-8'
-        
-        return text.encode(encoding, 'escapecss')
+            encoding = u'UTF-8'
+
+        # TODO: py3 return b str but tests use unicode?
+        return text.encode(encoding, u'escapecss')
 
     def do_CSSComment(self, rule):
         """
@@ -375,7 +404,33 @@ class CSSSerializer(object):
         no comments or other things allowed!
         """
         if rule.wellformed:
-            return u'@charset %s;' % self._string(rule.encoding)
+            return u'@charset %s;' % helper.string(rule.encoding)
+        else:
+            return u''
+
+    def do_CSSVariablesRule(self, rule):
+        """
+        serializes CSSVariablesRule
+
+        media
+            TODO
+        variables
+            CSSStyleDeclaration
+
+        + CSSComments
+        """
+        variablesText = rule.variables.cssText
+
+        if variablesText and rule.wellformed and not self.prefs.resolveVariables:
+            out = Out(self)
+            out.append(self._atkeyword(rule, u'@variables'))
+            for item in rule.seq:
+                # assume comments {
+                out.append(item.value, item.type)
+            out.append(u'{')
+            out.append(u'%s%s}' % (variablesText, self.prefs.lineSeparator),
+                       indent=1)
+            return out.value()
         else:
             return u''
 
@@ -392,14 +447,14 @@ class CSSSerializer(object):
 
         if styleText and rule.wellformed:
             out = Out(self)
-            out.append(self._atkeyword(rule, u'@font-face'))   
+            out.append(self._atkeyword(rule, u'@font-face'))
             for item in rule.seq:
                 # assume comments {
-                out.append(item.value, item.type)            
+                out.append(item.value, item.type)
             out.append(u'{')
             out.append(u'%s%s}' % (styleText, self.prefs.lineSeparator),
-                       indent=1)            
-            return out.value()            
+                       indent=1)
+            return out.value()
         else:
             return u''
 
@@ -412,14 +467,14 @@ class CSSSerializer(object):
         media
             optional cssutils.stylesheets.medialist.MediaList
         name
-            optional string 
+            optional string
 
         + CSSComments
         """
         if rule.wellformed:
             out = Out(self)
             out.append(self._atkeyword(rule, u'@import'))
-            
+
             for item in rule.seq:
                 typ, val = item.type, item.value
                 if 'href' == typ:
@@ -429,14 +484,12 @@ class CSSSerializer(object):
                              rule.hreftype == 'string'):
                         out.append(val, 'STRING')
                     else:
-                        if not len(self.prefs.spacer):
-                            out.append(u' ')   
                         out.append(val, 'URI')
                 elif 'media' == typ:
                     # media
                     mediaText = self.do_stylesheets_medialist(val)
                     if mediaText and mediaText != u'all':
-                        out.append(mediaText)                
+                        out.append(mediaText)
                 elif 'name' == typ:
                     out.append(val, 'STRING')
                 else:
@@ -460,20 +513,17 @@ class CSSSerializer(object):
         if rule.wellformed:
             out = Out(self)
             out.append(self._atkeyword(rule, u'@namespace'))
-            if not len(self.prefs.spacer):
-                out.append(u' ')   
-            
             for item in rule.seq:
                 typ, val = item.type, item.value
                 if 'namespaceURI' == typ:
                     out.append(val, 'STRING')
                 else:
                     out.append(val, typ)
-                    
+
             return out.value(end=u';')
         else:
             return u''
-        
+
     def do_CSSMediaRule(self, rule):
         """
         serializes CSSMediaRule
@@ -487,29 +537,29 @@ class CSSSerializer(object):
             return u''
 
         # @media
-        out = [self._atkeyword(rule, u'@media')] 
+        out = [self._atkeyword(rule, u'@media')]
         if not len(self.prefs.spacer):
             # for now always with space as only webkit supports @mediaall?
             out.append(u' ')
-        else: 
+        else:
             out.append(self.prefs.spacer) # might be empty
-        
+
         out.append(self.do_stylesheets_medialist(rule.media))
-        
+
         # name, seq contains content after name only (Comments)
         if rule.name:
             out.append(self.prefs.spacer)
             nameout = Out(self)
-            nameout.append(self._string(rule.name))
+            nameout.append(helper.string(rule.name))
             for item in rule.seq:
                 nameout.append(item.value, item.type)
             out.append(nameout.value())
-            
+
         #  {
         out.append(self.prefs.paranthesisSpacer)
-        out.append(u'{')                      
+        out.append(u'{')
         out.append(self.prefs.lineSeparator)
-        
+
         # rules
         rulesout = []
         for r in rule.cssRules:
@@ -519,12 +569,13 @@ class CSSSerializer(object):
                 rulesout.append(self._indentblock(rtext, self._level + 1))
                 rulesout.append(self.prefs.lineSeparator)
         if not self.prefs.keepEmptyRules and not u''.join(rulesout).strip():
-            return u''           
+            return u''
         out.extend(rulesout)
-        
+
         #     }
-        out.append(u'%s}' % ((self._level + 1) * self.prefs.indent))
-        
+        out.append(u'%s}' % ((self._level + int(self.prefs.indentClosingBrace)) 
+                             * self.prefs.indent))
+
         return u''.join(out)
 
     def do_CSSPageRule(self, rule):
@@ -539,35 +590,37 @@ class CSSSerializer(object):
         + CSSComments
         """
         styleText = self.do_css_CSSStyleDeclaration(rule.style)
-
         if styleText and rule.wellformed:
             out = Out(self)
             out.append(self._atkeyword(rule, u'@page'))
-            if not len(self.prefs.spacer):
-                out.append(u' ')
-                        
-            for item in rule.seq:
-                out.append(item.value, item.type)
-                            
-            out.append(u'{')            
+            out.append(rule.selectorText)
+            out.append(u'{')
             out.append(u'%s%s}' % (styleText, self.prefs.lineSeparator),
                        indent=1)
-            return out.value()            
+            return out.value()
         else:
             return u''
-        
+
+    def do_CSSPageRuleSelector(self, seq):
+        "Serialize selector of a CSSPageRule"
+        out = Out(self)
+        for item in seq:
+            if item.type == 'IDENT':
+                out.append(item.value, item.type, space=False)
+            else:
+                out.append(item.value, item.type)
+        return out.value()
+
     def do_CSSUnknownRule(self, rule):
         """
         serializes CSSUnknownRule
         anything until ";" or "{...}"
         + CSSComments
         """
-        if rule.wellformed:
+        if rule.wellformed and self.prefs.keepUnknownAtRules:
             out = Out(self)
-            out.append(rule.atkeyword)  
-            if not len(self.prefs.spacer):
-                out.append(u' ')
-                         
+            out.append(rule.atkeyword)
+
             stacks = []
             for item in rule.seq:
                 typ, val = item.type, item.value
@@ -577,7 +630,7 @@ class CSSSerializer(object):
                     stackblock = stacks.pop().value()
                     if stackblock:
                         val = self._indentblock(
-                               stackblock + self.prefs.lineSeparator + val, 
+                               stackblock + self.prefs.lineSeparator + val,
                                min(1, len(stacks)+1))
                     else:
                         val = self._indentblock(val, min(1, len(stacks)+1))
@@ -586,12 +639,12 @@ class CSSSerializer(object):
                     stacks[-1].append(val, typ)
                 else:
                     out.append(val, typ)
-                    
+
                 # POST
                 if u'{' == val:
                     # new stack level
                     stacks.append(Out(self))
-            
+
             return out.value()
         else:
             return u''
@@ -624,12 +677,12 @@ class CSSSerializer(object):
                 elif self._selectorlevel > 0:
                     self._selectorlevel -= 1
             else:
-                # save new reference                
+                # save new reference
                 self._selectors.append(rule.selectorList)
                 self._selectorlevel = 0
-        
+
         # TODO ^ RESOLVE!!!!
-        
+
         selectorText = self.do_css_SelectorList(rule.selectorList)
         if not selectorText or not rule.wellformed:
             return u''
@@ -651,14 +704,15 @@ class CSSSerializer(object):
                     self.prefs.lineSeparator,
                     self._indentblock(styleText, self._level + 1),
                     self.prefs.lineSeparator,
-                    (self._level + 1) * self.prefs.indent),
+                    (self._level + int(self.prefs.indentClosingBrace)) 
+                    * self.prefs.indent),
                 self._selectorlevel)
 
     def do_css_SelectorList(self, selectorlist):
         "comma-separated list of Selectors"
         # does not need Out() as it is too simple
         if selectorlist.wellformed:
-            out = [] 
+            out = []
             for part in selectorlist.seq:
                 if isinstance(part, cssutils.css.Selector):
                     out.append(part.selectorText)
@@ -668,13 +722,13 @@ class CSSSerializer(object):
             return sep.join(out)
         else:
             return u''
-                  
+
     def do_css_Selector(self, selector):
         """
         a single Selector including comments
-        
+
         an element has syntax (namespaceURI, name) where namespaceURI may be:
-        
+
         - cssutils._ANYNS => ``*|name``
         - None => ``name``
         - u'' => ``|name``
@@ -682,14 +736,14 @@ class CSSSerializer(object):
         """
         if selector.wellformed:
             out = Out(self)
-            
-            DEFAULTURI = selector._namespaces.get('', None)           
+
+            DEFAULTURI = selector._namespaces.get('', None)
             for item in selector.seq:
                 typ, val = item.type, item.value
-                if type(val) == tuple:
+                if isinstance(val, tuple):
                     # namespaceURI|name (element or attribute)
                     namespaceURI, name = val
-                    if DEFAULTURI == namespaceURI or (not DEFAULTURI and 
+                    if DEFAULTURI == namespaceURI or (not DEFAULTURI and
                                                       namespaceURI is None):
                         out.append(name, typ, space=False)
                     else:
@@ -701,22 +755,53 @@ class CSSSerializer(object):
                                                     namespaceURI)
                             except IndexError:
                                 prefix = u''
-                        
+
                         out.append(u'%s|%s' % (prefix, name), typ, space=False)
                 else:
                     out.append(val, typ, space=False, keepS=True)
-            
+
             return out.value()
-        else: 
+        else:
+            return u''
+
+    def do_css_CSSVariablesDeclaration(self, variables):
+        """Variables of CSSVariableRule."""
+        if len(variables.seq) > 0:
+            out = Out(self)
+
+            lastitem = len(variables.seq) - 1
+            for i, item in enumerate(variables.seq):
+                type_, val = item.type, item.value
+                if u'var' == type_:
+                    name, cssvalue = val
+                    if self.prefs.normalizedVarNames:
+                        name = normalize(name)
+                    out.append(name)
+                    out.append(u':')
+                    out.append(cssvalue.cssText)
+                    if i < lastitem or not self.prefs.omitLastSemicolon:
+                        out.append(u';')
+
+                elif isinstance(val, cssutils.css.CSSComment):
+                    # CSSComment
+                    out.append(val, 'COMMENT')
+                    out.append(self.prefs.lineSeparator)
+                else:
+                    out.append(val.cssText, type_)
+                    out.append(self.prefs.lineSeparator)
+
+            return out.value().strip()
+
+        else:
             return u''
 
     def do_css_CSSStyleDeclaration(self, style, separator=None):
         """
         Style declaration of CSSStyleRule
         """
-#        # TODO: use Out()
-        
-        # may be comments only       
+        # TODO: use Out()
+
+        # may be comments only
         if len(style.seq) > 0:
             if separator is None:
                 separator = self.prefs.lineSeparator
@@ -727,8 +812,8 @@ class CSSSerializer(object):
             else:
                 # only effective ones
                 _effective = style.getProperties()
-                seq = [item for item in style.seq 
-                         if (isinstance(item.value, cssutils.css.Property) 
+                seq = [item for item in style.seq
+                         if (isinstance(item.value, cssutils.css.Property)
                              and item.value in _effective)
                          or not isinstance(item.value, cssutils.css.Property)]
 
@@ -742,10 +827,11 @@ class CSSSerializer(object):
                         out.append(separator)
                 elif isinstance(val, cssutils.css.Property):
                     # PropertySimilarNameList
-                    out.append(self.do_Property(val))
-                    if not (self.prefs.omitLastSemicolon and i==len(seq)-1):
-                        out.append(u';')
-                    out.append(separator)
+                    if val.cssText:
+                        out.append(val.cssText)
+                        if not (self.prefs.omitLastSemicolon and i==len(seq)-1):
+                            out.append(u';')
+                        out.append(separator)
                 elif isinstance(val, cssutils.css.CSSUnknownRule):
                     # @rule
                     out.append(val.cssText)
@@ -767,14 +853,14 @@ class CSSSerializer(object):
         """
         Style declaration of CSSStyleRule
 
-        Property has a seqs attribute which contains seq lists for             
+        Property has a seqs attribute which contains seq lists for
         name, a CSSvalue and a seq list for priority
         """
         # TODO: use Out()
-        
+
         out = []
         if property.seqs[0] and property.wellformed and self._valid(property):
-            nameseq, cssvalue, priorityseq = property.seqs
+            nameseq, value, priorityseq = property.seqs
 
             #name
             for part in nameseq:
@@ -786,13 +872,13 @@ class CSSSerializer(object):
                     out.append(part)
 
             if out and (not property._mediaQuery or
-                        property._mediaQuery and cssvalue.cssText):
+                        property._mediaQuery and value.cssText):
                 # MediaQuery may consist of name only
                 out.append(u':')
                 out.append(self.prefs.propertyNameSpacer)
 
             # value
-            out.append(cssvalue.cssText)
+            out.append(value.cssText)
 
             # priority
             if out and priorityseq:
@@ -814,7 +900,6 @@ class CSSSerializer(object):
         a Properties priority "!" S* "important"
         """
         # TODO: use Out()
-        
         out = []
         for part in priorityseq:
             if hasattr(part, 'cssText'): # comments
@@ -825,31 +910,127 @@ class CSSSerializer(object):
                 out.append(part)
         return u''.join(out).strip()
 
-    def do_css_CSSValue(self, cssvalue):
-        """
-        serializes a CSSValue
-        """
-        # TODO: use Out()
-        # TODO: use self._valid(cssvalue)?
+    def do_css_PropertyValue(self, value, valuesOnly=False):
+        """Serializes a PropertyValue"""
+        if not value:
+            return u''
+        else:
+            out = Out(self)
+            for item in value.seq:
+                type_, val = item.type, item.value
+                if valuesOnly and type_ == cssutils.css.CSSComment:
+                    continue
+                elif hasattr(val, 'cssText'):
+                    # RGBColor or CSSValue if a CSSValueList
+                    out.append(val.cssText, type_)
+                else:
+                    if val and val[0] == val[-1] and val[0] in '\'"':
+                        val = helper.string(val[1:-1])
+                    # S must be kept! in between values but no extra space
+                    out.append(val, type_)
+
+            return out.value()
+
+    def do_css_Value(self, value, valuesOnly=None):
+        """Serializes a Value, valuesOnly is ignored"""
+        if not value:
+            return u''
+        else:
+            out = Out(self)
+            if value.type in (u'DIMENSION', u'NUMBER', u'PERCENTAGE'):
+                dim = value.dimension or u''
+                if value.value == 0:
+                    val = u'0'
+                    if value.dimension in ('cm', 'mm', 'in', 'px', 'pc', 'pt',
+                                           'em', 'ex'):
+                        dim = u''
+                elif value.value == int(value.value):
+                    # cut off after . which is zero anyway
+                    val = unicode(int(value.value))
+                elif self.prefs.omitLeadingZero and -1 < value.value < 1:
+                    v = unicode(value.value)
+                    val = v
+                    if value._sign == u'-':
+                        val = v[0] + v[2:]
+                    else:
+                        val = v[1:]
+                else:
+                    val = unicode(value.value)
+
+                # keep '+' if given
+                if value.value != 0 and value._sign == u'+':
+                    sign = u'+'
+                else:
+                    sign = u''
+                    
+                out.append(sign + val + dim, value.type)
         
+            else:
+                # e.g. URI
+                out.append(value.value, value.type)
+                
+        return out.value()            
+
+    def do_css_ColorValue(self, value, valuesOnly=False):
+        """Serialize a ColorValue, a HASH simple value or FUNCTION"""
+        try:
+            return {'FUNCTION': self.do_css_CSSFunction,
+                    'HASH': self.do_css_Value,
+                    'IDENT': self.do_css_Value
+                    }[value.colorType](value, 
+                                       valuesOnly=valuesOnly)
+        except KeyError, e:
+            return u''
+
+    def do_css_CSSFunction(self, cssvalue, valuesOnly=False):
+        """Serialize a CSS function value"""
         if not cssvalue:
             return u''
         else:
-            sep = u',%s' % self.prefs.listItemSpacer
-            out = []
-            for part in cssvalue.seq:
-                if hasattr(part, 'cssText'):
-                    # comments or CSSValue if a CSSValueList
-                    out.append(part.cssText)
-                elif isinstance(part, basestring) and part == u',':
-                    out.append(sep)
-                else:
-                    # TODO: escape func parameter if STRING!
-                    if part and part[0] == part[-1] and part[0] in '\'"':
-                        # string has " " around it in CSSValue!
-                        part = self._string(part[1:-1])
-                    out.append(part)
-            return (u''.join(out)).strip()
+            out = Out(self)
+            for item in cssvalue.seq:
+                type_, val = item.type, item.value
+                if valuesOnly and type_ == cssutils.css.CSSComment:
+                    continue
+                out.append(val, type_)
+            return out.value()
+
+    def do_css_MSValue(self, cssvalue, valuesOnly=False):
+        """Serialize an ExpressionValue (IE only),
+        should at least keep the original syntax"""
+        if not cssvalue:
+            return u''
+        else:
+            out = Out(self)
+            for item in cssvalue.seq:
+                type_, val = item.type, item.value
+                if valuesOnly and type_ == cssutils.css.CSSComment:
+                    continue
+                #val = self._possiblezero(cssvalue, type_, val)
+                # do no send type_ so no special cases!
+                out.append(val, None, space=False)
+
+            return out.value()
+
+    def do_css_CSSVariable(self, variable, IGNORED=False):
+        """Serializes a CSSVariable"""
+        if not variable or not variable.name:
+            return u''
+        else:
+            out = Out(self)
+            v = variable.value
+            if self.prefs.resolveVariables and v:
+                # resolve variable
+                out.append(v)
+
+            else:
+                # keep var(NAME)
+                out.append(u'var(', 'FUNCTION')
+                out.append(variable.name, 'IDENT')
+                out.append(u')')
+
+            return out.value()
+
 
     def do_stylesheets_medialist(self, medialist):
         """
