@@ -4,12 +4,15 @@
 """
 __all__ = ['Tokenizer', 'CSSProductions']
 __docformat__ = 'restructuredtext'
-__version__ = '$Id: tokenize2.py 1834 2009-08-02 12:20:21Z cthedot $'
+__version__ = '$Id$'
 
 from cssproductions import *
 from helper import normalize
+import itertools
 import re
 
+_TOKENIZER_CACHE = {}
+
 class Tokenizer(object):
     """
     generates a list of Token tuples:
@@ -18,28 +21,44 @@ class Tokenizer(object):
     _atkeywords = {
         u'@font-face': CSSProductions.FONT_FACE_SYM,
         u'@import': CSSProductions.IMPORT_SYM,
-        u'@media': CSSProductions.MEDIA_SYM,        
+        u'@media': CSSProductions.MEDIA_SYM,
         u'@namespace': CSSProductions.NAMESPACE_SYM,
-        u'@page': CSSProductions.PAGE_SYM
+        u'@page': CSSProductions.PAGE_SYM,
+        u'@variables': CSSProductions.VARIABLES_SYM
         }
     _linesep = u'\n'
-    unicodesub = re.compile(r'\\[0-9a-fA-F]{1,6}(?:\r\n|[\t|\r|\n|\f|\x20])?').sub
-    cleanstring = re.compile(r'\\((\r\n)|[\n|\r|\f])').sub
-    
-    def __init__(self, macros=None, productions=None):
+    unicodesub = re.compile(r'\\[0-9a-fA-F]{1,6}(?:\r\n|[\t\r\n\f\x20])?').sub
+    cleanstring = re.compile(r'\\((\r\n)|[\n\r\f])').sub
+
+    def __init__(self, macros=None, productions=None, doComments=True):
         """
         inits tokenizer with given macros and productions which default to
         cssutils own macros and productions
         """
-        if not macros:
-            macros = MACROS
-        if not productions:
-            productions = PRODUCTIONS
-        self.tokenmatches = self._compile_productions(
-                self._expand_macros(macros, 
-                                    productions))
-        self.commentmatcher = [x[1] for x in self.tokenmatches if x[0] == 'COMMENT'][0]
-        self.urimatcher = [x[1] for x in self.tokenmatches if x[0] == 'URI'][0]
+        if isinstance(macros, dict):
+            macros_hash_key = sorted(macros.items())
+        else:
+            macros_hash_key = macros
+        hash_key = str((macros_hash_key, productions))
+        if hash_key in _TOKENIZER_CACHE:
+            (tokenmatches, commentmatcher, urimatcher) = _TOKENIZER_CACHE[hash_key]
+        else:
+            if not macros:
+                macros = MACROS
+            if not productions:
+                productions = PRODUCTIONS
+            tokenmatches = self._compile_productions(self._expand_macros(macros,
+                                                                         productions))
+            commentmatcher = [x[1] for x in tokenmatches if x[0] == 'COMMENT'][0]
+            urimatcher = [x[1] for x in tokenmatches if x[0] == 'URI'][0]
+            _TOKENIZER_CACHE[hash_key] = (tokenmatches, commentmatcher, urimatcher)
+
+        self.tokenmatches = tokenmatches
+        self.commentmatcher = commentmatcher
+        self.urimatcher = urimatcher
+
+        self._doComments = doComments
+        self._pushed = []
 
     def _expand_macros(self, macros, productions):
         """returns macro expanded productions, order of productions is kept"""
@@ -60,13 +79,20 @@ class Tokenizer(object):
             compiled.append((key, re.compile('^(?:%s)' % value, re.U).match))
         return compiled
 
+    def push(self, *tokens):
+        """Push back tokens which have been pulled but not processed."""
+        self._pushed = itertools.chain(tokens, self._pushed)
+
+    def clear(self):
+        self._pushed = []
+
     def tokenize(self, text, fullsheet=False):
-        """Generator: Tokenize text and yield tokens, each token is a tuple 
+        """Generator: Tokenize text and yield tokens, each token is a tuple
         of::
-        
-            (nname, value, line, col)
-        
-        The token value will contain a normal string, meaning CSS unicode 
+
+            (name, value, line, col)
+
+        The token value will contain a normal string, meaning CSS unicode
         escapes have been resolved to normal characters. The serializer
         escapes needed characters back to unicode escapes depending on
         the stylesheet target encoding.
@@ -90,7 +116,7 @@ class Tokenizer(object):
             return normalize(self.unicodesub(_repl, value))
 
         line = col = 1
-        
+
         # check for BOM first as it should only be max one at the start
         (BOM, matcher), productions = self.tokenmatches[0], self.tokenmatches[1:]
         match = matcher(text)
@@ -105,69 +131,83 @@ class Tokenizer(object):
             yield (CSSProductions.CHARSET_SYM, found, line, col)
             text = text[len(found):]
             col += len(found)
-        
+
         while text:
-            # speed test for most used CHARs
+            # do pushed tokens before new ones
+            for pushed in self._pushed:
+                yield pushed
+
+            # speed test for most used CHARs, sadly . not possible :(
             c = text[0]
-            if c in '{}:;,':
+            if c in u',:;{}>+[]':
                 yield ('CHAR', c, line, col)
                 col += 1
                 text = text[1:]
-                
+
             else:
                 # check all other productions, at least CHAR must match
                 for name, matcher in productions:
+
+                    # TODO: USE bad comment?
                     if fullsheet and name == 'CHAR' and text.startswith(u'/*'):
                         # before CHAR production test for incomplete comment
                         possiblecomment = u'%s*/' % text
                         match = self.commentmatcher(possiblecomment)
-                        if match:
+                        if match and self._doComments:
                             yield ('COMMENT', possiblecomment, line, col)
-                            text = None # eats all remaining text 
-                            break 
-    
+                            text = None # ate all remaining text
+                            break
+
                     match = matcher(text) # if no match try next production
                     if match:
                         found = match.group(0) # needed later for line/col
-                        if fullsheet:                        
+                        if fullsheet:
                             # check if found may be completed into a full token
                             if 'INVALID' == name and text == found:
                                 # complete INVALID to STRING with start char " or '
                                 name, found = 'STRING', '%s%s' % (found, found[0])
-                            
+
                             elif 'FUNCTION' == name and\
                                  u'url(' == _normalize(found):
-                                # FUNCTION url( is fixed to URI if fullsheet
-                                # FUNCTION production MUST BE after URI production!
+                                # url( is a FUNCTION if incomplete sheet
+                                # FUNCTION production MUST BE after URI production
                                 for end in (u"')", u'")', u')'):
                                     possibleuri = '%s%s' % (text, end)
                                     match = self.urimatcher(possibleuri)
                                     if match:
                                         name, found = 'URI', match.group(0)
                                         break
-    
-                        if name in ('DIMENSION', 'IDENT', 'STRING', 'URI', 
+
+                        if name in ('DIMENSION', 'IDENT', 'STRING', 'URI',
                                     'HASH', 'COMMENT', 'FUNCTION', 'INVALID',
                                     'UNICODE-RANGE'):
-                            # may contain unicode escape, replace with normal char
-                            # but do not _normalize (?)
+                            # may contain unicode escape, replace with normal
+                            # char but do not _normalize (?)
                             value = self.unicodesub(_repl, found)
                             if name in ('STRING', 'INVALID'): #'URI'?
                                 # remove \ followed by nl (so escaped) from string
-                                value = self.cleanstring('', found)
-    
+                                value = self.cleanstring('', value)
+
                         else:
                             if 'ATKEYWORD' == name:
-                                # get actual ATKEYWORD SYM
-                                if '@charset' == found and ' ' == text[len(found):len(found)+1]:
-                                    # only this syntax!
-                                    name = CSSProductions.CHARSET_SYM
-                                    found += ' '
-                                else:
-                                    name = self._atkeywords.get(_normalize(found), 'ATKEYWORD')
-                                    
+                                try:
+                                    # get actual ATKEYWORD SYM
+                                    name = self._atkeywords[_normalize(found)]
+                                except KeyError, e:
+                                    # might also be misplace @charset...
+                                    if '@charset' == found and u' ' == text[len(found):len(found)+1]:
+                                        # @charset needs tailing S!
+                                        name = CSSProductions.CHARSET_SYM
+                                        found += u' '
+                                    else:
+                                        name = 'ATKEYWORD'
+
                             value = found # should not contain unicode escape (?)
-                        yield (name, value, line, col)
+
+                        if self._doComments or (not self._doComments and
+                                                name != 'COMMENT'):
+                            yield (name, value, line, col)
+
                         text = text[len(found):]
                         nls = found.count(self._linesep)
                         line += nls
@@ -175,6 +215,7 @@ class Tokenizer(object):
                             col = len(found[found.rfind(self._linesep):])
                         else:
                             col += len(found)
+
                         break
 
         if fullsheet:
